# Generated by Django 2.2.5 on 2019-11-04 16:28

from django.db import migrations, models

import asymmetric_jwt_auth.models


class Migration(migrations.Migration):
    dependencies = [
        ("asymmetric_jwt_auth", "0003_auto_20151112_1547"),
    ]

    operations = [
        migrations.AddField(
            model_name="publickey",
            name="last_used_on",
            field=models.DateTimeField(
                blank=True, null=True, verbose_name="Last Used On"
            ),
        ),
        migrations.AlterField(
            model_name="publickey",
            name="key",
            field=models.TextField(
                help_text="The user's RSA public key",
                validators=[asymmetric_jwt_auth.models.validate_public_key],
            ),
        ),
    ]



from django.urls import path
from django.views.decorators.cache import cache_page

from . import get_setting
from .views import JWKSEndpointView

_cache_jwks = cache_page(get_setting("JWKS_VIEW_TTL"))

app_name = "asymmetric_jwt_auth"
urlpatterns = [
    path(".well-known/jwks.json", _cache_jwks(JWKSEndpointView.as_view()), name="jwks"),
]



from typing import Union

from django.contrib.auth import get_user_model
from django.contrib.auth.models import User
from jwt.exceptions import PyJWKClientError

from .. import models
from ..tokens import Token, UntrustedToken
from .base import BasePublicKeyRepository, BaseUserRepository


class DjangoUserRepository(BaseUserRepository):
    def __init__(self) -> None:
        self.User = get_user_model()

    def get_user(self, username: str) -> Union[None, User]:
        """
        Get a Django user by username
        """
        try:
            return self.User.objects.get(username=username)
        except self.User.DoesNotExist:
            pass
        return None


class DjangoPublicKeyListRepository(BasePublicKeyRepository):
    def attempt_to_verify_token(
        self,
        user: User,
        untrusted_token: UntrustedToken,
    ) -> Union[Token, None]:
        """
        Attempt to verify a JWT for the given user using public keys from the PublicKey model.
        """
        for user_key in models.PublicKey.objects.filter(user=user).all():
            public_key = user_key.get_key()
            token = untrusted_token.verify(public_key=public_key)
            if token:
                user_key.update_last_used_datetime()
                return token
        return None


class DjangoJWKSRepository(BasePublicKeyRepository):
    def attempt_to_verify_token(
        self,
        user: User,
        untrusted_token: UntrustedToken,
    ) -> Union[Token, None]:
        """
        Attempt to verify a JWT for the given user using public keys the user's JWKS endpoint.
        """
        jwks_endpoints = models.JWKSEndpointTrust.objects.filter(user=user).all()
        for jwks_endpoint in jwks_endpoints:
            try:
                public_key = jwks_endpoint.get_signing_key(untrusted_token)
            except PyJWKClientError:
                continue
            token = untrusted_token.verify(public_key=public_key)
            if token:
                jwks_endpoint.update_last_used_datetime()
                return token
        return None



from typing import List

from django.utils.module_loading import import_string

from .. import get_setting
from .base import BasePublicKeyRepository, BaseUserRepository


def get_user_repository() -> BaseUserRepository:
    Repo: type[BaseUserRepository] = import_string(get_setting("USER_REPOSITORY"))
    return Repo()


def get_public_key_repositories() -> List[BasePublicKeyRepository]:
    repos = []
    for cls_path in get_setting("PUBLIC_KEY_REPOSITORIES"):
        Repo = import_string(cls_path)
        repo = Repo()
        repos.append(repo)
    return repos



from io import StringIO

from django.core.management import call_command
from django.test import TestCase


class ManagementCommandTest(TestCase):
    def test_generate_new_rsa_key(self):
        stdout = StringIO()
        stderr = StringIO()
        call_command("generate_key_pair", stdout=stdout, stderr=stderr)
        out = stdout.getvalue()
        self.assertIn("-----BEGIN PRIVATE KEY-----\n", out)
        self.assertIn("-----END PRIVATE KEY-----\n", out)
        self.assertIn("-----BEGIN PUBLIC KEY-----\n", out)
        self.assertIn("-----END PUBLIC KEY-----\n", out)
        self.assertTrue(len(out) > 2000)
        self.assertTrue(len(out) < 3000)

    def test_generate_new_ed25519_key(self):
        stdout = StringIO()
        stderr = StringIO()
        call_command(
            "generate_key_pair", keytype="Ed25519", stdout=stdout, stderr=stderr
        )
        out = stdout.getvalue()
        self.assertIn("-----BEGIN PRIVATE KEY-----\n", out)
        self.assertIn("-----END PRIVATE KEY-----\n", out)
        self.assertIn("-----BEGIN PUBLIC KEY-----\n", out)
        self.assertIn("-----END PUBLIC KEY-----\n", out)
        self.assertTrue(len(out) > 200)
        self.assertTrue(len(out) < 300)



from typing import Any, Generic, List, Optional, Tuple, TypeVar, Union
import hashlib
import os

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519, rsa

from .utils import long_to_base64

CryptoPrivateKey = Union[rsa.RSAPrivateKey, ed25519.Ed25519PrivateKey]
CryptoPublicKey = Union[rsa.RSAPublicKey, ed25519.Ed25519PublicKey]

FacadePrivateKey = Union["RSAPrivateKey", "Ed25519PrivateKey"]
FacadePublicKey = Union["RSAPublicKey", "Ed25519PublicKey"]

PrivateKeyType = TypeVar(
    "PrivateKeyType",
    bound=Union[rsa.RSAPrivateKey, ed25519.Ed25519PrivateKey],
)
PublicKeyType = TypeVar(
    "PublicKeyType",
    bound=Union[rsa.RSAPublicKey, ed25519.Ed25519PublicKey],
)


class PublicKey(Generic[PublicKeyType]):
    """Represents a public key"""

    _key: PublicKeyType

    @staticmethod
    def from_cryptography_pubkey(pubkey: CryptoPublicKey) -> FacadePublicKey:
        if isinstance(pubkey, rsa.RSAPublicKey):
            return RSAPublicKey(pubkey)
        if isinstance(pubkey, ed25519.Ed25519PublicKey):
            return Ed25519PublicKey(pubkey)
        raise TypeError(f"Unknown key type: {pubkey}")

    @classmethod
    def load_pem(cls, pem: bytes) -> FacadePublicKey:
        """
        Load a PEM-format public key
        """
        privkey = serialization.load_pem_public_key(pem)
        return cls.from_cryptography_pubkey(privkey)

    @classmethod
    def load_openssh(cls, key: bytes) -> FacadePublicKey:
        """
        Load a openssh-format public key
        """
        privkey = serialization.load_ssh_public_key(key)
        return cls.from_cryptography_pubkey(privkey)

    @classmethod
    def load_serialized_public_key(
        cls, key: bytes
    ) -> Tuple[Optional[Exception], Optional[FacadePublicKey]]:
        """
        Load a PEM or openssh format public key
        """
        exc = None
        for loader in (cls.load_pem, cls.load_openssh):
            try:
                return None, loader(key)
            except Exception as e:
                exc = e
        return exc, None

    @property
    def as_pem(self) -> bytes:
        """
        Get the public key as a PEM-formatted byte string
        """
        pem_bytes = self._key.public_bytes(
            serialization.Encoding.PEM, serialization.PublicFormat.SubjectPublicKeyInfo
        )
        return pem_bytes

    @property
    def as_jwk(self) -> dict[str, str]:  # pragma: no cover
        """
        Return the public key in JWK format
        """
        raise NotImplementedError("Subclass does not implement as_jwk method")

    @property
    def fingerprint(self) -> str:
        """
        Get a sha256 fingerprint of the key.
        """
        return hashlib.sha256(self.as_pem).hexdigest()

    @property
    def allowed_algorithms(self) -> List[str]:  # pragma: no cover
        """
        Return a list of allowed JWT algorithms for this key, in order of most to least preferred.
        """
        raise NotImplementedError(
            "Subclass does not implement allowed_algorithms method"
        )


class RSAPublicKey(PublicKey[rsa.RSAPublicKey]):
    """Represents an RSA public key"""

    def __init__(self, key: rsa.RSAPublicKey):
        self._key = key

    @property
    def as_jwk(self) -> dict[str, str]:
        """
        Return the public key in JWK format
        """
        public_numbers = self._key.public_numbers()
        return {
            "kty": "RSA",
            "use": "sig",
            "alg": self.allowed_algorithms[0],
            "kid": self.fingerprint,
            "n": long_to_base64(public_numbers.n),
            "e": long_to_base64(public_numbers.e),
        }

    @property
    def allowed_algorithms(self) -> List[str]:
        return [
            "RS512",
            "RS384",
            "RS256",
        ]


class Ed25519PublicKey(PublicKey[ed25519.Ed25519PublicKey]):
    """Represents an Ed25519 public key"""

    def __init__(self, key: ed25519.Ed25519PublicKey):
        self._key = key

    @property
    def allowed_algorithms(self) -> List[str]:
        return [
            "EdDSA",
        ]


class PrivateKey(Generic[PrivateKeyType]):
    """Represents a private key"""

    _key: PrivateKeyType

    @staticmethod
    def from_cryptography_privkey(privkey: CryptoPrivateKey) -> FacadePrivateKey:
        if isinstance(privkey, rsa.RSAPrivateKey):
            return RSAPrivateKey(privkey)
        if isinstance(privkey, ed25519.Ed25519PrivateKey):
            return Ed25519PrivateKey(privkey)
        raise TypeError("Unknown key type")

    @classmethod
    def load_pem_from_file(
        cls,
        filepath: os.PathLike[Any],
        password: Optional[bytes] = None,
    ) -> FacadePrivateKey:
        """
        Load a PEM-format private key from disk.
        """
        with open(filepath, "rb") as fh:
            key_bytes = fh.read()
        return cls.load_pem(key_bytes, password=password)

    @classmethod
    def load_pem(cls, pem: bytes, password: Optional[bytes] = None) -> FacadePrivateKey:
        """
        Load a PEM-format private key
        """
        privkey = serialization.load_pem_private_key(pem, password=password)
        return cls.from_cryptography_privkey(privkey)

    @property
    def as_pem(self) -> bytes:
        pem_bytes = self._key.private_bytes(  # type:ignore[union-attr]
            serialization.Encoding.PEM,
            serialization.PrivateFormat.PKCS8,
            serialization.NoEncryption(),
        )
        return pem_bytes

    @property
    def public_key(self) -> FacadePublicKey:  # pragma: no cover
        raise NotImplementedError()


class RSAPrivateKey(PrivateKey[rsa.RSAPrivateKey]):
    """Represents an RSA private key"""

    pubkey_cls = RSAPublicKey

    @classmethod
    def generate(
        cls, size: int = 2048, public_exponent: int = 65537
    ) -> "RSAPrivateKey":
        """
        Generate an RSA private key.
        """
        private = rsa.generate_private_key(
            public_exponent=public_exponent, key_size=size
        )
        return cls(private)

    def __init__(self, key: rsa.RSAPrivateKey):
        self._key = key

    @property
    def public_key(self) -> FacadePublicKey:
        public = self._key.public_key()
        return self.pubkey_cls(public)


class Ed25519PrivateKey(PrivateKey[ed25519.Ed25519PrivateKey]):
    """Represents an Ed25519 private key"""

    pubkey_cls = Ed25519PublicKey

    @classmethod
    def generate(cls) -> "Ed25519PrivateKey":
        """
        Generate an Ed25519 private key.
        """
        private = ed25519.Ed25519PrivateKey.generate()
        return cls(private)

    def __init__(self, key: ed25519.Ed25519PrivateKey):
        self._key = key

    @property
    def public_key(self) -> FacadePublicKey:
        public = self._key.public_key()
        return self.pubkey_cls(public)



import json

from django.core.cache import cache
from django.test import Client, TestCase, override_settings
from django.urls import reverse

from ..keys import PublicKey
from . import data


class JWKSViewTest(TestCase):
    def setUp(self):
        cache.clear()

    def test_no_keys(self):
        client = Client()
        response = client.get(reverse("asymmetric_jwt_auth:jwks"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            json.loads(response.content),
            {
                "keys": [],
            },
        )

    @override_settings(
        ASYMMETRIC_JWT_AUTH=dict(
            SIGNING_PUBLIC_KEYS=[data.PEM_PUBLIC_RSA, data.PEM_PUBLIC_RSA],
        )
    )
    def test_pem_keys(self):
        client = Client()
        response = client.get(reverse("asymmetric_jwt_auth:jwks"))
        self.assertEqual(response.status_code, 200)
        jwk = {
            "alg": "RS512",
            "e": "AQAB",
            "kid": "53c5b68c5ecba3e25df3f8326de6c0b0befb67e9217651a2f40e388f6567f056",
            "kty": "RSA",
            "n": "odxbRh5LOtoB3Shf6K3mRn7ME7Doo5Qm5h72ITt-E6U0l6qXGdVBTj0XhQVNnGjnZTGzU7IacIw1a_03qVHJfcc0Ki7ig7YSPMMl0WSp0m080YlsCZ-9g-WG6DrgjpGQU7yaBhNsKtR5DP20bm8411S9VLqV2GEOzBKpB10_lwhRZuv_Qj7obwSqdVCzMNb7t5LHqG0MxOF7BeYELXIqTEKFfWkZytXCAnmC9hk9RtzUZ_lryD1UgCHZ16gPtmPdFV7fuN8FBNrbaQCldz6V6HVDjsPVxPmVYswV8qInG8kJUpm48s9PAWfgi4HCGmJgn-Irbed2tlRf73jxyCgX0Q",  # NOQA
            "use": "sig",
        }
        self.assertEqual(
            json.loads(response.content),
            {
                "keys": [
                    jwk,
                    jwk,
                ],
            },
        )

    @override_settings(
        ASYMMETRIC_JWT_AUTH=dict(
            SIGNING_PUBLIC_KEYS=[
                PublicKey.load_pem(data.PEM_PUBLIC_RSA),
            ],
        )
    )
    def test_loaded_keys(self):
        client = Client()
        response = client.get(reverse("asymmetric_jwt_auth:jwks"))
        self.assertEqual(response.status_code, 200)
        jwk = {
            "alg": "RS512",
            "e": "AQAB",
            "kid": "53c5b68c5ecba3e25df3f8326de6c0b0befb67e9217651a2f40e388f6567f056",
            "kty": "RSA",
            "n": "odxbRh5LOtoB3Shf6K3mRn7ME7Doo5Qm5h72ITt-E6U0l6qXGdVBTj0XhQVNnGjnZTGzU7IacIw1a_03qVHJfcc0Ki7ig7YSPMMl0WSp0m080YlsCZ-9g-WG6DrgjpGQU7yaBhNsKtR5DP20bm8411S9VLqV2GEOzBKpB10_lwhRZuv_Qj7obwSqdVCzMNb7t5LHqG0MxOF7BeYELXIqTEKFfWkZytXCAnmC9hk9RtzUZ_lryD1UgCHZ16gPtmPdFV7fuN8FBNrbaQCldz6V6HVDjsPVxPmVYswV8qInG8kJUpm48s9PAWfgi4HCGmJgn-Irbed2tlRf73jxyCgX0Q",  # NOQA
            "use": "sig",
        }
        self.assertEqual(
            json.loads(response.content),
            {
                "keys": [
                    jwk,
                ],
            },
        )

    @override_settings(
        ASYMMETRIC_JWT_AUTH=dict(
            SIGNING_PUBLIC_KEYS=[
                data.PEM_PUBLIC_RSA_INVALID,
            ],
        )
    )
    def test_invalid_pem_keys(self):
        client = Client()
        response = client.get(reverse("asymmetric_jwt_auth:jwks"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            json.loads(response.content),
            {
                "keys": [],
            },
        )



# Generated by Django 5.1.4 on 2025-02-05 16:43

from django.db import migrations, models

import asymmetric_jwt_auth.models


class Migration(migrations.Migration):

    dependencies = [
        ("asymmetric_jwt_auth", "0005_auto_20210304_1116"),
    ]

    operations = [
        migrations.AddField(
            model_name="jwksendpointtrust",
            name="last_used_on",
            field=models.DateTimeField(
                blank=True, null=True, verbose_name="Last Used On"
            ),
        ),
        migrations.AlterField(
            model_name="publickey",
            name="key",
            field=models.TextField(
                help_text="The user's RSA/Ed25519 public key",
                validators=[asymmetric_jwt_auth.models.validate_public_key],
                verbose_name="Public Key",
            ),
        ),
    ]



from typing import Any
import copy

from django.conf import settings

#: Default settings. Override using a dictionary named ASYMMETRIC_JWT_AUTH in Django's settings.py.
default_settings = {
    #: Auth method searched for in the prefix of the Authentication header. Similar to ``Bearer`` or ``Basic``.
    "AUTH_METHOD": "JWT",
    #: Number of seconds of clock-drift to tolerate when verifying the authenticity of a JWT.
    "TIMESTAMP_TOLERANCE": 20,  # Seconds
    #: Class used to store and validate nonces
    "NONCE_BACKEND": "asymmetric_jwt_auth.nonce.django.DjangoCacheNonceBackend",
    #: Repository class used to fetch users by their username
    "USER_REPOSITORY": "asymmetric_jwt_auth.repos.django.DjangoUserRepository",
    #: List of repository classes used to fetch public keys for a user
    "PUBLIC_KEY_REPOSITORIES": [
        "asymmetric_jwt_auth.repos.django.DjangoJWKSRepository",
        "asymmetric_jwt_auth.repos.django.DjangoPublicKeyListRepository",
    ],
    #: List of public keys that should be advertised on our JWKS endpoint.
    "SIGNING_PUBLIC_KEYS": [],
    #: Cache TTL for the JWKS key view
    "JWKS_VIEW_TTL": (60 * 5),
}


def get_setting(name: str) -> Any:
    _settings = copy.deepcopy(default_settings)
    _settings.update(getattr(settings, "ASYMMETRIC_JWT_AUTH", {}))
    return _settings[name]



from django.utils.module_loading import import_string

from .. import get_setting
from .base import BaseNonceBackend


def get_nonce_backend() -> BaseNonceBackend:
    backend_path = get_setting("NONCE_BACKEND")
    Backend: type[BaseNonceBackend] = import_string(backend_path)
    return Backend()



from django.http import HttpRequest, JsonResponse
from django.views import View

from . import get_setting
from .keys import FacadePublicKey, PublicKey


class JWKSEndpointView(View):
    def get(self, request: HttpRequest) -> JsonResponse:
        keys = self.list_pub_keys()
        data = {
            "keys": [key.as_jwk for key in keys],
        }
        return JsonResponse(data)

    def list_pub_keys(self) -> list[FacadePublicKey]:
        keys: list[FacadePublicKey] = []
        for _key in get_setting("SIGNING_PUBLIC_KEYS"):
            if isinstance(_key, PublicKey):
                keys.append(_key)  # type:ignore[arg-type]
            else:
                exc, key = PublicKey.load_serialized_public_key(_key)
                if key is not None:
                    keys.append(key)
        return keys



from functools import lru_cache
from typing import Any

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone
from django.utils.encoding import force_bytes, force_str
from django.utils.translation import gettext_lazy as _
from jwt import PyJWKClient

from . import keys, tokens


def validate_public_key(keystr: str) -> None:
    """
    Check that the given value is a valid public key in either PEM or OpenSSH format. If it is invalid,
    raises ``django.core.exceptions.ValidationError``.
    """
    key_bytes = keystr.encode()
    exc, key = keys.PublicKey.load_serialized_public_key(key_bytes)
    is_valid = (exc is None) and (key is not None)
    if not is_valid:
        raise ValidationError("Public key is invalid: %s" % exc)


@lru_cache
def get_jwks_client(jwks_url: str) -> PyJWKClient:
    """
    LRU-cached constructor for PyJWKClient. Re-using instances of this (scoped
    by the URL of course) allows the JWKS caching inside PyJWKClient to function
    correctly.
    """
    return PyJWKClient(jwks_url)


class PublicKey(models.Model):
    """
    Store a public key and associate it to a particular user.

    Implements the same concept as the OpenSSH ``~/.ssh/authorized_keys`` file on a Unix system.
    """

    #: Foreign key to the Django User model. Related name: ``public_keys``.
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("User"),
        related_name="public_keys",
        on_delete=models.CASCADE,
    )

    #: Key text in either PEM or OpenSSH format.
    key = models.TextField(
        _("Public Key"),
        help_text=_("The user's RSA/Ed25519 public key"),
        validators=[validate_public_key],
    )

    #: Comment describing the key. Use this to note what system is authenticating with the key, when it was last rotated, etc.
    comment = models.CharField(
        _("Comment"),
        max_length=100,
        help_text=_("Comment describing this key"),
        blank=True,
    )

    #: Date and time that key was last used for authenticating a request.
    last_used_on = models.DateTimeField(_("Last Used On"), null=True, blank=True)

    class Meta:
        verbose_name = _("Public Key")
        verbose_name_plural = _("Public Keys")

    def get_key(self) -> keys.FacadePublicKey:
        key_bytes = force_bytes(self.key)
        exc, key = keys.PublicKey.load_serialized_public_key(key_bytes)
        if key is None:
            if exc is None:  # pragma: no cover
                raise ValueError("Failed to load key")
            raise exc
        return key

    def update_last_used_datetime(self) -> None:
        self.last_used_on = timezone.now()
        self.save(update_fields=["last_used_on"])

    def save(self, *args: Any, **kwargs: Any) -> None:
        key_parts = force_str(self.key).split(" ")
        if len(key_parts) == 3 and not self.comment:
            self.comment = key_parts.pop()
        super(PublicKey, self).save(*args, **kwargs)


class JWKSEndpointTrust(models.Model):
    """
    Associate a JSON Web Key Set (JWKS) URL with a Django User.

    This accomplishes the same purpose of the PublicKey model, in a more automated
    fashion. Instead of manually assigning a public key to a user, the system will
    load a list of public keys from this URL.
    """

    #: Foreign key to the Django User model. Related name: ``public_keys``.
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        verbose_name=_("User"),
        related_name="jwks_endpoint",
        on_delete=models.CASCADE,
    )

    #: URL of the JSON Web Key Set (JWKS)
    jwks_url = models.URLField(
        _("JSON Web Key Set (JWKS) URL"),
        help_text=_("e.g. https://dev-87evx9ru.auth0.com/.well-known/jwks.json"),
    )

    #: Date and time that key was last used for authenticating a request.
    last_used_on = models.DateTimeField(_("Last Used On"), null=True, blank=True)

    class Meta:
        verbose_name = _("JSON Web Key Set")
        verbose_name_plural = _("JSON Web Key Sets")

    @property
    def jwks_client(self) -> PyJWKClient:
        return get_jwks_client(self.jwks_url)

    def get_signing_key(
        self, untrusted_token: tokens.UntrustedToken
    ) -> keys.FacadePublicKey:
        jwk = self.jwks_client.get_signing_key_from_jwt(untrusted_token.token)
        return keys.PublicKey.from_cryptography_pubkey(jwk.key)

    def update_last_used_datetime(self) -> None:
        self.last_used_on = timezone.now()
        self.save(update_fields=["last_used_on"])



from unittest import mock
import os.path

from django.test import TestCase

from .. import keys
from . import data


class PublicKeyTest(TestCase):
    def test_load_pem_rsa(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.PEM_PUBLIC_RSA)
        self.assertIsNone(exc)
        self.assertIsInstance(key, keys.RSAPublicKey)

    def test_load_pem_ed25519(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.PEM_PUBLIC_ED25519)
        self.assertIsNone(exc)
        self.assertIsInstance(key, keys.Ed25519PublicKey)

    def test_load_openssh_rsa(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.OPENSSH_RSA)
        self.assertIsNone(exc)
        self.assertIsInstance(key, keys.RSAPublicKey)

    def test_load_openssh_ed25519(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.OPENSSH_ED25519)
        self.assertIsNone(exc)
        self.assertIsInstance(key, keys.Ed25519PublicKey)

    def test_load_invalid_pem_rsa(self):
        exc, key = keys.PublicKey.load_serialized_public_key(
            data.PEM_PUBLIC_RSA_INVALID
        )
        self.assertIsInstance(exc, Exception)
        self.assertIsNone(key)

    def test_load_invalid_pem_ed25519(self):
        exc, key = keys.PublicKey.load_serialized_public_key(
            data.PEM_PUBLIC_ED25519_INVALID
        )
        self.assertIsInstance(exc, Exception)
        self.assertIsNone(key)

    def test_load_invalid_openssh_rsa(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.OPENSSH_RSA_INVALID)
        self.assertIsInstance(exc, Exception)
        self.assertIsNone(key)

    def test_load_invalid_openssh_ed25519(self):
        exc, key = keys.PublicKey.load_serialized_public_key(
            data.OPENSSH_ED25519_INVALID
        )
        self.assertIsInstance(exc, Exception)
        self.assertIsNone(key)

    def test_rsa_as_pem(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.PEM_PUBLIC_RSA)
        self.assertIsNone(exc)
        pem = key.as_pem.decode().strip().split("\n")
        self.assertEqual(pem[0], "-----BEGIN PUBLIC KEY-----")
        self.assertEqual(pem[-1], "-----END PUBLIC KEY-----")

    def test_ed25519_as_pem(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.PEM_PUBLIC_RSA)
        self.assertIsNone(exc)
        pem = key.as_pem.decode().strip().split("\n")
        self.assertEqual(pem[0], "-----BEGIN PUBLIC KEY-----")
        self.assertEqual(pem[-1], "-----END PUBLIC KEY-----")

    def test_rsa_fingerprint(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.PEM_PUBLIC_RSA)
        self.assertIsNone(exc)
        self.assertEqual(
            key.fingerprint,
            "53c5b68c5ecba3e25df3f8326de6c0b0befb67e9217651a2f40e388f6567f056",
        )

    def test_ed25519_fingerprint(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.PEM_PUBLIC_ED25519)
        self.assertIsNone(exc)
        self.assertEqual(
            key.fingerprint,
            "cb10cd75c2eacf7aa2b5195bef9838cccd9d2ae4938601178808cb881b68ec72",
        )

    def test_rsa_allowed_algorithms(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.PEM_PUBLIC_RSA)
        self.assertIsNone(exc)
        self.assertEqual(
            key.allowed_algorithms,
            [
                "RS512",
                "RS384",
                "RS256",
            ],
        )

    def test_ed25519_allowed_algorithms(self):
        exc, key = keys.PublicKey.load_serialized_public_key(data.PEM_PUBLIC_ED25519)
        self.assertIsNone(exc)
        self.assertEqual(
            key.allowed_algorithms,
            [
                "EdDSA",
            ],
        )

    def test_unknown_key_type(self):
        with self.assertRaises(TypeError):
            keys.PublicKey.from_cryptography_pubkey(mock.MagicMock())


class PrivateKeyTest(TestCase):
    def test_generate_rsa(self):
        private = keys.RSAPrivateKey.generate()
        self.assertIsInstance(private, keys.RSAPrivateKey)
        self.assertIsInstance(private.public_key, keys.RSAPublicKey)

    def test_generate_ed25519(self):
        private = keys.Ed25519PrivateKey.generate()
        self.assertIsInstance(private, keys.Ed25519PrivateKey)
        self.assertIsInstance(private.public_key, keys.Ed25519PublicKey)

    def test_load_rsa_from_file(self):
        base = os.path.dirname(__file__)
        filepath = os.path.join(base, "fixtures/dummy_rsa.privkey")
        private = keys.RSAPrivateKey.load_pem_from_file(filepath)
        self.assertIsInstance(private, keys.RSAPrivateKey)

    def test_load_rsa_from_file_encrypted(self):
        base = os.path.dirname(__file__)
        filepath = os.path.join(base, "fixtures/dummy_rsa_encrypted.privkey")
        private = keys.RSAPrivateKey.load_pem_from_file(filepath, password=b"password")
        self.assertIsInstance(private, keys.RSAPrivateKey)

    def test_load_rsa(self):
        priv1 = keys.RSAPrivateKey.generate()
        priv2 = keys.RSAPrivateKey.load_pem(priv1.as_pem)
        self.assertEqual(priv2.as_pem, priv1.as_pem)

    def test_load_ed25519(self):
        priv1 = keys.Ed25519PrivateKey.generate()
        priv2 = keys.Ed25519PrivateKey.load_pem(priv1.as_pem)
        self.assertEqual(priv2.as_pem, priv1.as_pem)

    def test_rsa_as_pem(self):
        private = keys.RSAPrivateKey.generate()
        pem = private.as_pem.decode().strip().split("\n")
        self.assertEqual(pem[0], "-----BEGIN PRIVATE KEY-----")
        self.assertEqual(pem[-1], "-----END PRIVATE KEY-----")

    def test_ed25519_as_pem(self):
        private = keys.Ed25519PrivateKey.generate()
        pem = private.as_pem.decode().strip().split("\n")
        self.assertEqual(pem[0], "-----BEGIN PRIVATE KEY-----")
        self.assertEqual(pem[-1], "-----END PRIVATE KEY-----")

    def test_unknown_key_type(self):
        with self.assertRaises(TypeError):
            keys.PrivateKey.from_cryptography_privkey(mock.MagicMock())



from unittest import mock

from django.contrib.auth.models import User
from django.test import RequestFactory, TestCase

from ..keys import Ed25519PrivateKey, RSAPrivateKey
from ..middleware import JWTAuthMiddleware
from ..models import JWKSEndpointTrust, PublicKey
from ..tokens import Token


class BaseMiddlewareTest(TestCase):
    def assertNotLoggedIn(self, request):
        self.assertEqual(getattr(request, "user", None), None)

    def assertLoggedIn(self, request, public_key=None):
        if public_key:
            public_key.refresh_from_db()
            self.assertIsNotNone(public_key.last_used_on)
        self.assertEqual(getattr(request, "user", None), self.user)


class MiddlewareTest(BaseMiddlewareTest):
    def setUp(self):
        self.rfactory = RequestFactory()
        self.user = User.objects.create_user(username="foo")
        self.user2 = User.objects.create_user(username="bar")

        self.key_ed25519 = Ed25519PrivateKey.generate()
        self.key_rsa = RSAPrivateKey.generate()

        self.user_key_ed25519 = PublicKey.objects.create(
            user=self.user, key=self.key_ed25519.public_key.as_pem.decode()
        )
        self.user_key_rsa = PublicKey.objects.create(
            user=self.user, key=self.key_rsa.public_key.as_pem.decode()
        )

        self.next_middleware = mock.MagicMock()
        self.run_middleware = JWTAuthMiddleware(self.next_middleware)

    def test_no_auth_header(self):
        request = self.rfactory.get("/")
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertNotLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)

    def test_auth_header_missing_type(self):
        request = self.rfactory.get("/", HTTP_AUTHORIZATION="Fooopbar")
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertNotLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)

    def test_auth_header_not_jwt_type(self):
        request = self.rfactory.get("/", HTTP_AUTHORIZATION="Bearer foobar")
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertNotLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)

    def test_header_jwt_missing_username(self):
        header = Token("").create_auth_header(self.key_rsa)
        request = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertNotLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)

    def test_header_jwt_claimed_username_doesnt_exist(self):
        header = Token("rusty").create_auth_header(self.key_rsa)
        request = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertNotLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)

    def test_authenticate_request_rsa_valid(self):
        header = Token(self.user.username).create_auth_header(self.key_rsa)
        request = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertLoggedIn(request, self.user_key_rsa)
        self.assertEqual(self.next_middleware.call_count, 1)

    def test_authenticate_request_rsa_unregistered_key(self):
        # Assign the pub keys to user 2
        self.user_key_ed25519.user = self.user2
        self.user_key_ed25519.save()
        self.user_key_rsa.user = self.user2
        self.user_key_rsa.save()
        # Try to use user2's key to login as user1
        header = Token(self.user.username).create_auth_header(self.key_rsa)
        request = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertNotLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)

    def test_authenticate_request_ed25519_valid(self):
        header = Token(self.user.username).create_auth_header(self.key_ed25519)
        request = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertLoggedIn(request, self.user_key_ed25519)
        self.assertEqual(self.next_middleware.call_count, 1)

    def test_missing_data(self):
        header = Token(self.user.username, timestamp=0).create_auth_header(
            self.key_ed25519
        )
        request = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertNotLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)

    def test_cant_reuse_nonce(self):
        header = Token(self.user.username).create_auth_header(self.key_ed25519)
        # First use works
        request1 = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request1)
        self.run_middleware(request1)
        self.assertLoggedIn(request1, self.user_key_ed25519)
        self.assertEqual(self.next_middleware.call_count, 1)
        # Second use doesn't
        request2 = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request2)
        self.run_middleware(request2)
        self.assertNotLoggedIn(request2)
        self.assertEqual(self.next_middleware.call_count, 2)


class MiddlewareJWKSTest(BaseMiddlewareTest):
    def setUp(self):
        self.rfactory = RequestFactory()
        self.user = User.objects.create_user(username="foo")

        self.key_ed25519 = Ed25519PrivateKey.generate()
        self.key_rsa = RSAPrivateKey.generate()

        self.next_middleware = mock.MagicMock()
        self.run_middleware = JWTAuthMiddleware(self.next_middleware)

    @mock.patch("asymmetric_jwt_auth.models.PyJWKClient.fetch_data")
    def test_authenticate_request_rsa(self, mock_fetch_data):
        mock_fetch_data.return_value = {
            "keys": [
                self.key_rsa.public_key.as_jwk,
            ],
        }
        JWKSEndpointTrust.objects.create(user=self.user, jwks_url="")
        header = Token(self.user.username).create_auth_header(self.key_rsa)
        request = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)

    @mock.patch("asymmetric_jwt_auth.models.PyJWKClient.fetch_data")
    def test_no_matching_key(self, mock_fetch_data):
        # Change the key ID of the JWKS response so it doesn't match the kid in the header JWT
        jwk = self.key_rsa.public_key.as_jwk
        jwk["kid"] = "foobar"
        mock_fetch_data.return_value = {
            "keys": [
                jwk,
            ],
        }
        JWKSEndpointTrust.objects.create(user=self.user, jwks_url="")
        header = Token(self.user.username).create_auth_header(self.key_rsa)
        request = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertNotLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)

    @mock.patch("asymmetric_jwt_auth.models.PyJWKClient.fetch_data")
    def test_header_jwt_claimed_username_doesnt_exist(self, mock_fetch_data):
        mock_fetch_data.return_value = {
            "keys": [
                self.key_rsa.public_key.as_jwk,
            ],
        }
        JWKSEndpointTrust.objects.create(user=self.user, jwks_url="")
        header = Token("rusty").create_auth_header(self.key_rsa)
        request = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertNotLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)

    @mock.patch("asymmetric_jwt_auth.models.PyJWKClient.fetch_data")
    def test_missing_data(self, mock_fetch_data):
        mock_fetch_data.return_value = {
            "keys": [
                self.key_rsa.public_key.as_jwk,
            ],
        }
        JWKSEndpointTrust.objects.create(user=self.user, jwks_url="")
        header = Token(self.user.username, timestamp=0).create_auth_header(self.key_rsa)
        request = self.rfactory.get("/", HTTP_AUTHORIZATION=header)
        self.assertNotLoggedIn(request)
        self.run_middleware(request)
        self.assertNotLoggedIn(request)
        self.assertEqual(self.next_middleware.call_count, 1)



from django.contrib import admin

from . import models


@admin.register(models.PublicKey)
class PublicKeyAdmin(admin.ModelAdmin):  # type:ignore[type-arg]
    list_display = ["user", "comment", "last_used_on"]
    fields = ["user", "comment", "key", "last_used_on"]
    raw_id_fields = ["user"]


@admin.register(models.JWKSEndpointTrust)
class JWKSEndpointTrustAdmin(admin.ModelAdmin):  # type:ignore[type-arg]
    list_display = ["user", "jwks_url", "last_used_on"]
    fields = ["user", "jwks_url", "last_used_on"]
    raw_id_fields = ["user"]



from typing import List
import base64
import struct


def long2intarr(long_int: int) -> List[int]:
    _bytes: List[int] = []
    while long_int:
        long_int, r = divmod(long_int, 256)
        _bytes.insert(0, r)
    return _bytes


def long_to_base64(n: int, mlen: int = 0) -> str:
    bys = long2intarr(n)
    if mlen:
        _len = mlen - len(bys)
        if _len:
            bys = [0] * _len + bys
    data = struct.pack("%sB" % len(bys), *bys)
    if not len(data):
        data = b"\x00"
    s = base64.urlsafe_b64encode(data).rstrip(b"=")
    return s.decode("ascii")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("asymmetric_jwt_auth", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="publickey",
            name="comment",
            field=models.CharField(
                max_length=100, help_text="Comment describing this key", default=""
            ),
            preserve_default=False,
        ),
    ]



PEM_PUBLIC_RSA = b"""
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAodxbRh5LOtoB3Shf6K3m
Rn7ME7Doo5Qm5h72ITt+E6U0l6qXGdVBTj0XhQVNnGjnZTGzU7IacIw1a/03qVHJ
fcc0Ki7ig7YSPMMl0WSp0m080YlsCZ+9g+WG6DrgjpGQU7yaBhNsKtR5DP20bm84
11S9VLqV2GEOzBKpB10/lwhRZuv/Qj7obwSqdVCzMNb7t5LHqG0MxOF7BeYELXIq
TEKFfWkZytXCAnmC9hk9RtzUZ/lryD1UgCHZ16gPtmPdFV7fuN8FBNrbaQCldz6V
6HVDjsPVxPmVYswV8qInG8kJUpm48s9PAWfgi4HCGmJgn+Irbed2tlRf73jxyCgX
0QIDAQAB
-----END PUBLIC KEY-----
"""

PEM_PUBLIC_ED25519 = b"""
-----BEGIN PUBLIC KEY-----
MCowBQYDK2VwAyEAhRk96LXVjEtq8yI1I5LiRiv0OHiGvgJKfU0a4SweOe0=
-----END PUBLIC KEY-----
"""

PEM_PUBLIC_RSA_INVALID = b"""
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAodxbRh5LOtoB3Shf6K3m
Rn7ME7Doo5Qm5h72ITt+E6U0l6qXGdVBTj0XhQVNnGjnZTGzU7IacIw1a/03qVHJ
11S9VLqV2GEOzBKpB10/lwhRZuv/Qj7obwSqdVCzMNb7t5LHqG0MxOF7BeYELXIq
TEKFfWkZytXCAnmC9hk9RtzUZ/lryD1UgCHZ16gPtmPdFV7fuN8FBNrbaQCldz6V
6HVDjsPVxPmVYswV8qInG8kJUpm48s9PAWfgi4HCGmJgn+Irbed2tlRf73jxyCgX
0QIDAQAB
-----END PUBLIC KEY-----
"""

PEM_PUBLIC_ED25519_INVALID = b"""
-----BEGIN PUBLIC KEY-----
MCowBQYDK2VwAyEAhRk96LXVjEtq8yI1I5Lv0OHiGvgJKfU0a4SweOe0=
-----END PUBLIC KEY-----
"""

OPENSSH_RSA = b"ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAACAQChfejJdi6Jbg4ealsjfC8Jwy3ucwU7PcLWDEVhEi+rvgLRmWhIbK1Tt8lOGx2JECu6zymbFpYSH7cacUqpZfp/Bm4LMtFLqxXqeXymsGmH5mAJYqd0jKZtk0IM8RAvbn9iUvWtmqYXDcE734+dhvsfPEu8LDP251TIskslbj8XIKwQd4q1ervNmhG7o6culFSTltsLwDQ5LdopRfp2cu5i3umNXKBpbYcYDfa4YASmTra/rF+cp9YMXQkTTpsGBRn9wOnJmxRpFEdJ0QDBDqL4zAHkY0Fc4GJufl/4HoYmkolYxzkiYku6wd8bDMcU+o4XZ/78eNoYLPrjCHHy0ytPtFDZMuYB+e8DLGkVp3lNGfV+BRX+s/bexrBRLZoA9U2B7YHq7BOaZs8VRFehU/q0AICM0AOqKHFX3dJPKtEEUb4wmeFS/MoZQm2DXHIhkOA64A+ltdklGgHEjy8daQBvjJ0yIx5IfPMGFpZgk8/ETRcqHTEmmbU1ri6CevQrM7PFCGnmk3btFYUDUHTgykaTr9IA2W+yTMLwKKXBpJlr8lA4oRQpaNpdkuwUY9ivWtTycpl0v5YwLFYsJPcFQPJD31G8AXXBp58K/0YXlt2SuA+kg4QAlFHmJdOAfs8LeQLD01fWhlIWFJlLRS1NHKOOvWKT8YM8kx76I6Ck861Dxw== crgwbr@foo"  # NOQA

OPENSSH_ED25519 = b"ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIDNkE30ChofcWQbPLrWhR+7uJkwEtRO2UCI2WxRiRpU3 crgwbr@foo"

OPENSSH_RSA_INVALID = b"ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAACAQChfejJLRmWhIbK1Tt8lOGx2JECu6zymbFpYSH7cacUqpZfp/Bm4LMtFLqxXqeXymsGmH5mAJYqd0jKZtk0IM8RAvbn9iUvWtmqYXDcE734+dhvsfPEu8LDP251TIskslbj8XIKwQd4q1ervNmhG7o6culFSTltsLwDQ5LdopRfp2cu5i3umNXKBpbYcYDfa4YASmTra/rF+cp9YMXQkTTpsGBRn9wOnJmxRpFEdJ0QDBDqL4zAHkY0Fc4GJufl/4HoYmkolYxzkiYku6wd8bDMcU+o4XZ/78eNoYLPrjCHHy0ytPtFDZMuYB+e8DLGkVp3lNGfV+BRX+s/bexrBRLZoA9U2B7YHq7BOaZs8VRFehU/q0AICM0AOqKHFX3dJPKtEEUb4wmeFS/MoZQm2DXHIhkOA64A+ltdklGgHEjy8daQBvjJ0yIx5IfPMGFpZgk8/ETRcqHTEmmbU1ri6CevQrM7PFCGnmk3btFYUDUHTgykaTr9IA2W+yTMLwKKXBpJlr8lA4oRQpaNpdkuwUY9ivWtTycpl0v5YwLFYsJPcFQPJD31G8AXXBp58K/0YXlt2SuA+kg4QAlFHmJdOAfs8LeQLD01fWhlIWFJlLRS1NHKOOvWKT8YM8kx76I6Ck861Dxw== crgwbr@foo"  # NOQA

OPENSSH_ED25519_INVALID = b"ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIDNkE30ChohR+7uJkwEtRO2UCI2WxRiRpU3 crgwbr@foo"



from argparse import ArgumentParser
from typing import Any

from django.core.management.base import BaseCommand

from asymmetric_jwt_auth.keys import Ed25519PrivateKey, FacadePrivateKey, RSAPrivateKey

TYPE_RSA = "RSA"
TYPE_ED25519 = "Ed25519"
TYPE_CHOICES = [
    TYPE_RSA,
    TYPE_ED25519,
]


class Command(BaseCommand):
    help = "Generate a public / private RSA key pair"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-t", "--keytype", choices=TYPE_CHOICES, default=TYPE_RSA)

    def handle(self, *args: Any, keytype: str = TYPE_RSA, **options: Any) -> None:
        privkey: FacadePrivateKey
        if keytype == TYPE_ED25519:
            privkey = Ed25519PrivateKey.generate()
        else:
            privkey = RSAPrivateKey.generate()
        self.stdout.write(privkey.as_pem.decode())
        self.stdout.write(privkey.public_key.as_pem.decode())



from typing import Optional, Union
import logging
import time

import jwt

from . import get_setting, keys
from .nonce import get_nonce_backend

logger = logging.getLogger(__name__)


class Token:
    """
    Represents a JWT that's either been constructed by our code or has been
    verified to be valid.
    """

    username: str
    timestamp: int

    def __init__(self, username: str, timestamp: Optional[int] = None):
        self.username = username
        self.timestamp = int(time.time()) if timestamp is None else timestamp

    def create_auth_header(self, private_key: keys.FacadePrivateKey) -> str:
        """
        Create an HTTP Authorization header
        """

        auth_method = get_setting("AUTH_METHOD")
        token = self.sign(private_key)
        return f"{auth_method} {token}"

    def sign(self, private_key: keys.FacadePrivateKey) -> str:
        """
        Create and return signed authentication JWT
        """
        public_key = private_key.public_key
        algorithm = public_key.allowed_algorithms[0]
        nonce = get_nonce_backend().generate_nonce()
        kid = public_key.fingerprint
        # Build and sign claim data
        token_data = {
            "username": self.username,
            "time": self.timestamp,
            "nonce": nonce,
        }
        headers = {
            "kid": kid,
        }
        token = jwt.encode(
            payload=token_data,
            key=private_key.as_pem,
            algorithm=algorithm,
            headers=headers,
        )
        return token


class UntrustedToken:
    """
    Represents a JWT received from user input (and not yet trusted)
    """

    token: str

    def __init__(self, token: str):
        self.token = token

    def get_claimed_username(self) -> Union[None, str]:
        """
        Given a JWT, get the username that it is claiming to be `without verifying that the signature is valid`.

        :param token: JWT claim
        :return: Username
        """
        unverified_data = jwt.decode(self.token, options={"verify_signature": False})
        username = unverified_data.get("username")
        if isinstance(username, str):
            return username
        return None

    def verify(self, public_key: keys.FacadePublicKey) -> Union[None, Token]:
        """
        Verify the validity of the given JWT using the given public key.
        """
        try:
            token_data = jwt.decode(
                jwt=self.token,
                key=public_key.as_pem.decode(),
                algorithms=public_key.allowed_algorithms,
            )
        except jwt.InvalidTokenError:
            logger.debug("JWT failed verification")
            return None

        claimed_username = token_data.get("username")
        claimed_time = token_data.get("time", 0)
        claimed_nonce = token_data.get("nonce")

        # Ensure fields aren't blank
        if not claimed_username or not claimed_time or not claimed_nonce:
            return None

        # Ensure time is within acceptable bounds
        current_time = time.time()
        timestamp_tolerance = get_setting("TIMESTAMP_TOLERANCE")
        min_time, max_time = (
            current_time - timestamp_tolerance,
            current_time + timestamp_tolerance,
        )
        if claimed_time < min_time or claimed_time > max_time:
            logger.debug("Claimed time is outside of allowable tolerances")
            return None

        # Ensure nonce is unique
        nonce_backend = get_nonce_backend()
        if not nonce_backend.validate_nonce(
            claimed_username, claimed_time, claimed_nonce
        ):
            logger.debug("Claimed nonce failed to validate")
            return None

        # If we've gotten this far, the token is valid
        nonce_backend.log_used_nonce(claimed_username, claimed_time, claimed_nonce)
        return Token(claimed_username, claimed_time)



from django.test import TestCase, override_settings

from .. import get_setting


class SettingsTest(TestCase):
    def test_get_settings_default(self):
        self.assertEqual(get_setting("AUTH_METHOD"), "JWT")
        self.assertEqual(get_setting("TIMESTAMP_TOLERANCE"), 20)

    @override_settings(ASYMMETRIC_JWT_AUTH=dict(TIMESTAMP_TOLERANCE=30))
    def test_get_setting_overridden(self):
        self.assertEqual(get_setting("AUTH_METHOD"), "JWT")
        self.assertEqual(get_setting("TIMESTAMP_TOLERANCE"), 30)



from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone
import jwt

from .. import keys, models, tokens
from . import data


class ValidatePublicKeyTest(TestCase):
    def test_valid_rsa_pem(self):
        models.validate_public_key(data.PEM_PUBLIC_RSA.decode())

    def test_valid_ed25519_pem(self):
        models.validate_public_key(data.PEM_PUBLIC_ED25519.decode())

    def test_valid_rsa_openssh(self):
        models.validate_public_key(data.OPENSSH_RSA.decode())

    def test_valid_ed25519_openssh(self):
        models.validate_public_key(data.OPENSSH_ED25519.decode())

    def test_invalid_rsa_pem(self):
        with self.assertRaises(ValidationError):
            models.validate_public_key(data.PEM_PUBLIC_RSA_INVALID.decode())

    def test_invalid_ed25519_pem(self):
        with self.assertRaises(ValidationError):
            models.validate_public_key(data.PEM_PUBLIC_ED25519_INVALID.decode())

    def test_invalid_rsa_openssh(self):
        with self.assertRaises(ValidationError):
            models.validate_public_key(data.OPENSSH_RSA_INVALID.decode())

    def test_invalid_ed25519_openssh(self):
        with self.assertRaises(ValidationError):
            models.validate_public_key(data.OPENSSH_ED25519_INVALID.decode())


class PublicKeyTest(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(username="foo")

    def test_extract_comment(self):
        pub = models.PublicKey(user=self.user, key=data.OPENSSH_ED25519, comment="")
        pub.save()
        self.assertEqual(pub.comment, "crgwbr@foo")

    def test_update_last_used_datetime(self):
        pub = models.PublicKey(user=self.user, key=data.OPENSSH_ED25519)
        pub.save()
        self.assertEqual(pub.last_used_on, None)
        pub.update_last_used_datetime()
        # Check the first 19 digits (year – second precision) of ISO time: 2021-03-03T17:00:24
        self.assertEqual(
            pub.last_used_on.isoformat()[:19], timezone.now().isoformat()[:19]
        )

    def test_get_key_ed25519(self):
        pub = models.PublicKey(user=self.user, key=data.OPENSSH_ED25519)
        pub.save()
        self.assertIsInstance(pub.get_key(), keys.Ed25519PublicKey)

    def test_get_loaded_key_rsa(self):
        pub = models.PublicKey(user=self.user, key=data.OPENSSH_RSA, comment="")
        pub.save()
        self.assertIsInstance(pub.get_key(), keys.RSAPublicKey)

    def test_get_loaded_key_invalid(self):
        pub = models.PublicKey(user=self.user, key=data.OPENSSH_ED25519_INVALID)
        pub.save()
        with self.assertRaises(ValueError):
            pub.get_key()


class JWKSEndpointTrustTest(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(username="foo")
        self.jwks = models.JWKSEndpointTrust.objects.create(
            user=self.user,
            jwks_url="https://dev-87evx9ru.auth0.com/.well-known/jwks.json",
        )

    def test_get_signing_key(self):
        untrusted_token = tokens.UntrustedToken(
            b"eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiIsImtpZCI6Ik5FRTFRVVJCT1RNNE16STVSa0ZETlRZeE9UVTFNRGcyT0Rnd1EwVXpNVGsxUWpZeVJrUkZRdyJ9.eyJpc3MiOiJodHRwczovL2Rldi04N2V2eDlydS5hdXRoMC5jb20vIiwic3ViIjoiYVc0Q2NhNzl4UmVMV1V6MGFFMkg2a0QwTzNjWEJWdENAY2xpZW50cyIsImF1ZCI6Imh0dHBzOi8vZXhwZW5zZXMtYXBpIiwiaWF0IjoxNTcyMDA2OTU0LCJleHAiOjE1NzIwMDY5NjQsImF6cCI6ImFXNENjYTc5eFJlTFdVejBhRTJINmtEME8zY1hCVnRDIiwiZ3R5IjoiY2xpZW50LWNyZWRlbnRpYWxzIn0.PUxE7xn52aTCohGiWoSdMBZGiYAHwE5FYie0Y1qUT68IHSTXwXVd6hn02HTah6epvHHVKA2FqcFZ4GGv5VTHEvYpeggiiZMgbxFrmTEY0csL6VNkX1eaJGcuehwQCRBKRLL3zKmA5IKGy5GeUnIbpPHLHDxr-GXvgFzsdsyWlVQvPX2xjeaQ217r2PtxDeqjlf66UYl6oY6AqNS8DH3iryCvIfCcybRZkc_hdy-6ZMoKT6Piijvk_aXdm7-QQqKJFHLuEqrVSOuBqqiNfVrG27QzAPuPOxvfXTVLXL2jek5meH6n-VWgrBdoMFH93QEszEDowDAEhQPHVs0xj7SIzA"  # NOQA
        )
        signing_key = self.jwks.get_signing_key(untrusted_token)
        self.assertIsInstance(signing_key, keys.RSAPublicKey)
        data = jwt.decode(
            untrusted_token.token,
            signing_key.as_pem,
            algorithms=["RS256"],
            audience="https://expenses-api",
            options={"verify_exp": False},
        )
        self.assertEqual(
            data,
            {
                "iss": "https://dev-87evx9ru.auth0.com/",
                "sub": "aW4Cca79xReLWUz0aE2H6kD0O3cXBVtC@clients",
                "aud": "https://expenses-api",
                "iat": 1572006954,
                "exp": 1572006964,
                "azp": "aW4Cca79xReLWUz0aE2H6kD0O3cXBVtC",
                "gty": "client-credentials",
            },
        )



from . import BaseNonceBackend


class NullNonceBackend(BaseNonceBackend):
    """
    Nonce backend which doesn't actually do anything
    """

    def validate_nonce(self, username: str, timestamp: int, nonce: str) -> bool:
        """
        Confirm that the given nonce hasn't already been used.
        """
        return True

    def log_used_nonce(self, username: str, timestamp: int, nonce: str) -> None:
        """
        Log a nonce as being used, and therefore henceforth invalid.
        """
        pass



from datetime import datetime, timedelta
from unittest.mock import patch

from django.contrib.auth.models import User
from django.test import TestCase, override_settings
from freezegun import freeze_time
import jwt

from .. import keys, tokens
from . import data


class TokenToken(TestCase):
    def setUp(self):
        self.username = "rusty"
        self.privkey = keys.Ed25519PrivateKey.generate()

    @patch("time.time")
    @patch("secrets.token_urlsafe")
    def test_create_auth_header(self, mock_get_nonce, mock_time):
        mock_get_nonce.return_value = "yVJ0MVWhqPQ"
        mock_time.return_value = 1234000.3
        token = tokens.Token(username=self.username)
        header = token.create_auth_header(self.privkey)
        self.assertTrue(header.startswith("JWT "))
        data = jwt.decode(
            jwt=header.split(" ")[1],
            key=self.privkey.public_key.as_pem,
            algorithms=self.privkey.public_key.allowed_algorithms,
        )
        self.assertEqual(
            data,
            {
                "nonce": "yVJ0MVWhqPQ",
                "time": 1234000,
                "username": "rusty",
            },
        )

    @patch("secrets.token_urlsafe")
    def test_create_auth_header_custom_time(self, mock_get_nonce):
        mock_get_nonce.return_value = "yVJ0MVWhqPQ"
        token = tokens.Token(username=self.username, timestamp=1614974974)
        header = token.create_auth_header(self.privkey)
        self.assertTrue(header.startswith("JWT "))
        data = jwt.decode(
            jwt=header.split(" ")[1],
            key=self.privkey.public_key.as_pem,
            algorithms=self.privkey.public_key.allowed_algorithms,
        )
        self.assertEqual(
            data,
            {
                "nonce": "yVJ0MVWhqPQ",
                "time": 1614974974,
                "username": "rusty",
            },
        )


class UntrustedTokenTest(TestCase):
    def setUp(self):
        self.username = "rusty"
        self.user = User.objects.create(username=self.username)
        self.privkey = keys.Ed25519PrivateKey.generate()
        self.token = tokens.Token(username=self.username)
        self.jwt_value = self.token.sign(self.privkey)
        self.untrusted_token = tokens.UntrustedToken(self.jwt_value)

    def test_get_claimed_username(self):
        self.assertEqual(self.untrusted_token.get_claimed_username(), self.username)

    def test_verify_valid(self):
        token = self.untrusted_token.verify(self.privkey.public_key)
        self.assertIsInstance(token, tokens.Token)
        self.assertEqual(token.username, self.username)

    def test_verify_key_mismatch(self):
        pubkey = keys.PublicKey.load_pem(data.PEM_PUBLIC_RSA)
        token = self.untrusted_token.verify(pubkey)
        self.assertIsNone(token)

    def test_time_out_of_allowed_range_before(self):
        dt = datetime.now() - timedelta(seconds=30)
        with freeze_time(dt):
            token = self.untrusted_token.verify(self.privkey.public_key)
        self.assertIsNone(token)

    def test_time_out_of_allowed_range_after(self):
        dt = datetime.now() + timedelta(seconds=30)
        with freeze_time(dt):
            token = self.untrusted_token.verify(self.privkey.public_key)
        self.assertIsNone(token)

    def test_nonce_already_used(self):
        token1 = self.untrusted_token.verify(self.privkey.public_key)
        self.assertIsInstance(token1, tokens.Token)
        self.assertEqual(token1.username, self.username)
        # Second attempt fails because nonce was already used
        token2 = self.untrusted_token.verify(self.privkey.public_key)
        self.assertIsNone(token2)

    @override_settings(
        ASYMMETRIC_JWT_AUTH=dict(
            NONCE_BACKEND="asymmetric_jwt_auth.nonce.null.NullNonceBackend"
        )
    )
    def test_nonce_already_used_null_backend(self):
        token1 = self.untrusted_token.verify(self.privkey.public_key)
        self.assertIsInstance(token1, tokens.Token)
        self.assertEqual(token1.username, self.username)
        # Second attempt works becuase null nonce backend doesn't do anything
        token2 = self.untrusted_token.verify(self.privkey.public_key)
        self.assertIsInstance(token2, tokens.Token)
        self.assertEqual(token2.username, self.username)



from typing import Union

from django.contrib.auth.models import User

from ..tokens import Token, UntrustedToken


class BaseUserRepository:
    def get_user(self, username: str) -> Union[None, User]:  # pragma: no cover
        raise NotImplementedError()


class BasePublicKeyRepository:
    def attempt_to_verify_token(
        self, user: User, untrusted_token: UntrustedToken
    ) -> Union[Token, None]:  # pragma: no cover
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="PublicKey",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        auto_created=True,
                        verbose_name="ID",
                        primary_key=True,
                    ),
                ),
                ("key", models.TextField(help_text="The user's RSA public key")),
                (
                    "user",
                    models.ForeignKey(
                        to=settings.AUTH_USER_MODEL,
                        related_name="public_keys",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 3.1.7 on 2021-03-04 11:16

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion

import asymmetric_jwt_auth.models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("asymmetric_jwt_auth", "0004_auto_20191104_1628"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="publickey",
            options={
                "verbose_name": "Public Key",
                "verbose_name_plural": "Public Keys",
            },
        ),
        migrations.AlterField(
            model_name="publickey",
            name="comment",
            field=models.CharField(
                blank=True,
                help_text="Comment describing this key",
                max_length=100,
                verbose_name="Comment",
            ),
        ),
        migrations.AlterField(
            model_name="publickey",
            name="key",
            field=models.TextField(
                help_text="The user's RSA public key",
                validators=[asymmetric_jwt_auth.models.validate_public_key],
                verbose_name="Public Key",
            ),
        ),
        migrations.AlterField(
            model_name="publickey",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="public_keys",
                to=settings.AUTH_USER_MODEL,
                verbose_name="User",
            ),
        ),
        migrations.CreateModel(
            name="JWKSEndpointTrust",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "jwks_url",
                    models.URLField(
                        help_text="e.g. https://dev-87evx9ru.auth0.com/.well-known/jwks.json",
                        verbose_name="JSON Web Key Set (JWKS) URL",
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="jwks_endpoint",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="User",
                    ),
                ),
            ],
            options={
                "verbose_name": "JSON Web Key Set",
                "verbose_name_plural": "JSON Web Key Sets",
            },
        ),
    ]



from typing import Callable
import logging

from django.http import HttpRequest, HttpResponse

from . import get_setting
from .nonce import get_nonce_backend
from .repos import get_public_key_repositories, get_user_repository
from .tokens import UntrustedToken

logger = logging.getLogger(__name__)


class JWTAuthMiddleware:
    """Django middleware class for authenticating users using JWT Authentication headers"""

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]):
        self.get_response = get_response
        self.nonce_backend = get_nonce_backend()
        self.user_repo = get_user_repository()
        self.key_repos = get_public_key_repositories()

    def __call__(self, request: HttpRequest) -> HttpResponse:
        # Attempt to authorize the request
        self.authorize_request(request)
        # Continue with the request
        return self.get_response(request)

    def authorize_request(self, request: HttpRequest) -> HttpRequest:
        """
        Process a Django request and authenticate users.

        If a JWT authentication header is detected and it is determined to be valid, the user is set as
        ``request.user`` and CSRF protection is disabled (``request._dont_enforce_csrf_checks = True``) on
        the request.

        :param request: Django Request instance
        """
        # Check for presence of auth header
        if "HTTP_AUTHORIZATION" not in request.META:
            return request

        # Ensure this auth header was meant for us (it has the JWT auth method).
        try:
            method, header_data = request.META["HTTP_AUTHORIZATION"].split(" ", 1)
        except ValueError:
            return request

        auth_method_setting = get_setting("AUTH_METHOD")
        if method.upper() != auth_method_setting:
            return request

        # Get the (unvalidated!) username that the request is claiming to be
        untrusted_token = UntrustedToken(header_data)
        username = untrusted_token.get_claimed_username()
        if not username:
            return request

        # Get the user model
        user = self.user_repo.get_user(username=username)
        if not user:
            return request

        # Try and validate the token using a key from the key repo
        verified_token = None
        for repo in self.key_repos:
            verified_token = repo.attempt_to_verify_token(user, untrusted_token)
            if verified_token:
                break

        # No keys successfully validated the claim? Abort.
        if not verified_token:
            return request

        # Assign the user to the request
        logger.debug("Successfully authenticated %s using JWT", user.username)
        request._dont_enforce_csrf_checks = True  # type:ignore[attr-defined]
        request.user = user
        return request



from django.conf import settings
from django.core.cache import cache

from .. import default_settings
from . import BaseNonceBackend


class DjangoCacheNonceBackend(BaseNonceBackend):
    """
    Nonce backend which uses DJango's cache system.

    Simple, but not great. Prone to race conditions.
    """

    def validate_nonce(self, username: str, timestamp: int, nonce: str) -> bool:
        """
        Confirm that the given nonce hasn't already been used.
        """
        key = self._create_nonce_key(username, timestamp)
        used = cache.get(key, set([]))
        return nonce not in used

    def log_used_nonce(self, username: str, timestamp: int, nonce: str) -> None:
        """
        Log a nonce as being used, and therefore henceforth invalid.
        """
        key = self._create_nonce_key(username, timestamp)
        used = cache.get(key, set([]))
        used.add(nonce)
        timestamp_tolerance = getattr(
            settings, "ASYMMETRIC_JWT_AUTH", default_settings
        )["TIMESTAMP_TOLERANCE"]
        cache.set(key, used, timestamp_tolerance * 2)

    def _create_nonce_key(self, username: str, timestamp: int) -> str:
        """
        Create and return the cache key for storing nonces
        """
        return "%s-nonces-%s-%s" % (
            self.__class__.__name__,
            username,
            timestamp,
        )



import secrets


class BaseNonceBackend:
    def validate_nonce(
        self, username: str, timestamp: int, nonce: str
    ) -> bool:  # pragma: no cover
        raise NotImplementedError()

    def log_used_nonce(
        self, username: str, timestamp: int, nonce: str
    ) -> None:  # pragma: no cover
        raise NotImplementedError()

    def generate_nonce(self) -> str:
        return secrets.token_urlsafe(nbytes=8)
